@@ -1,4 +1,4 @@
-#     Copyright 2014, Kay Hayen, mailto:kay.hayen@gmail.com
+#     Copyright 2015, Kay Hayen, mailto:kay.hayen@gmail.com
 #
 #     Part of "Nuitka", an optimizing Python compiler that is compatible and
 #     integrates with CPython, but also works on its own.
@@ -22,10 +22,12 @@ of operations allowed.
 
 """
 
-from .Utils import python_version
-
 import operator
 
+from nuitka.utils.Utils import python_version
+
+# TODO: Monkey patching this may cause issues for compile time assessments of
+# user code. This should be solved differently.
 if python_version >= 300:
     operator.div = operator.truediv
     operator.idiv = operator.itruediv
@@ -59,6 +61,11 @@ binary_operator_functions = {
     "IBitXor"   : operator.ixor,
 }
 
+# Python 3.5 only operator
+if python_version >= 350:
+    binary_operator_functions["MatMult"] = operator.matmul  # @UndefinedVariable
+    binary_operator_functions["IMatMult"] = operator.imatmul  # @UndefinedVariable
+
 unary_operator_functions = {
     "UAdd"   : operator.pos,
     "USub"   : operator.neg,
@@ -92,5 +99,24 @@ comparison_inversions = {
     "NotIn" : "In"
 }
 
-all_comparison_functions = dict( rich_comparison_functions)
-all_comparison_functions.update( other_comparison_functions )
+all_comparison_functions = dict(rich_comparison_functions)
+all_comparison_functions.update(other_comparison_functions)
+
+def matchException(left, right):
+    if python_version >= 300:
+        if type(right) is tuple:
+            for element in right:
+                if not isinstance(BaseException,element):
+                    raise TypeError("catching classes that do not inherit from BaseException is not allowed")
+        elif not isinstance(BaseException,right):
+            raise TypeError("catching classes that do not inherit from BaseException is not allowed")
+
+    # This doesn't yet work, make it error exit. and silence PyLint for now.
+    # pylint: disable=W0212
+    import os
+    os._exit(16)
+
+    assert False, left
+
+
+all_comparison_functions["exception_match"]=matchException
