@@ -1,4 +1,4 @@
-#     Copyright 2014, Kay Hayen, mailto:kay.hayen@gmail.com
+#     Copyright 2015, Kay Hayen, mailto:kay.hayen@gmail.com
 #
 #     Part of "Nuitka", an optimizing Python compiler that is compatible and
 #     integrates with CPython, but also works on its own.
@@ -23,16 +23,14 @@ result of it can be predicted still, and these are interesting for warnings.
 
 """
 
-from .NodeBases import (
-    ExpressionChildrenHavingBase,
-    ExpressionBuiltinNoArgBase
-)
+import math
 
 from nuitka.optimizations import BuiltinOptimization
+from nuitka.utils.Utils import python_version
 
-from nuitka.Utils import python_version
+from .NodeBases import ExpressionBuiltinNoArgBase, ExpressionChildrenHavingBase
+from .NodeMakingHelpers import makeConstantReplacementNode
 
-import math
 
 class ExpressionBuiltinRange0(ExpressionBuiltinNoArgBase):
     kind = "EXPRESSION_BUILTIN_RANGE0"
@@ -47,6 +45,9 @@ class ExpressionBuiltinRange0(ExpressionBuiltinNoArgBase):
     def mayHaveSideEffects(self):
         return False
 
+    def mayBeNone(self):
+        return False
+
 
 class ExpressionBuiltinRangeBase(ExpressionChildrenHavingBase):
     """ Base class for range nodes with 1/2/3 arguments. """
@@ -80,23 +81,47 @@ class ExpressionBuiltinRangeBase(ExpressionChildrenHavingBase):
                child.isExpressionConstantRef() and \
                type(child.getConstant()) is float:
                 return True
-        else:
-            return False
 
-    def computeBuiltinSpec(self, given_values):
+        return False
+
+    def mayRaiseException(self, exception_type):
+        for child in self.getVisitableNodes():
+            if child.mayRaiseException(exception_type):
+                return True
+
+            # TODO: Should take exception_type value into account here.
+            if child.getIntegerValue() is None:
+                return True
+
+            if python_version >= 270 and \
+               child.isExpressionConstantRef() and \
+               type(child.getConstant()) is float:
+                return True
+
+        step = self.getStep()
+
+        # A step of 0 will raise.
+        if step is not None and child.getIntegerValue() == 0:
+            return True
+
+        return False
+
+    def computeBuiltinSpec(self, constraint_collection, given_values):
         assert self.builtin_spec is not None, self
 
         if not self.builtin_spec.isCompileTimeComputable(given_values):
+            constraint_collection.onExceptionRaiseExit(BaseException)
+
+            # TODO: Raise exception known step 0.
+
             return self, None, None
 
-        from .NodeMakingHelpers import getComputationResult
-
-        return getComputationResult(
+        return constraint_collection.getCompileTimeComputationResult(
             node        = self,
             computation = lambda : self.builtin_spec.simulateCall(
                 given_values
             ),
-            description = "Builtin call to %s precomputed." % (
+            description = "Built-in call to '%s' computed." % (
                 self.builtin_spec.getName()
             )
         )
@@ -118,21 +143,38 @@ class ExpressionBuiltinRangeBase(ExpressionChildrenHavingBase):
 
             self.replaceWith(result)
 
-            return iter_node, "new_expression", "Replaced range with xrange."
+            return (
+                iter_node,
+                "new_expression",
+                "Replaced 'range' with 'xrange' built-in call."
+            )
+
+        # No exception will be raised on ranges.
 
         return iter_node, None, None
 
-    def getHigh(self):
+    @staticmethod
+    def getLow():
         return None
 
-    def getStep(self):
+    @staticmethod
+    def getHigh():
         return None
 
+    @staticmethod
+    def getStep():
+        return None
+
+    def mayBeNone(self):
+        return False
+
 
 class ExpressionBuiltinRange1(ExpressionBuiltinRangeBase):
     kind = "EXPRESSION_BUILTIN_RANGE1"
 
-    named_children = ( "low", )
+    named_children = (
+        "low",
+    )
 
     def __init__(self, low, source_ref):
         assert low is not None
@@ -140,16 +182,14 @@ class ExpressionBuiltinRange1(ExpressionBuiltinRangeBase):
         ExpressionBuiltinRangeBase.__init__(
             self,
             values     = {
-                "low"  : low,
+                "low" : low,
             },
             source_ref = source_ref
         )
 
-    getLow = ExpressionChildrenHavingBase.childGetter( "low" )
+    getLow = ExpressionChildrenHavingBase.childGetter("low")
 
     def computeExpression(self, constraint_collection):
-        # Children can tell all we need to know, pylint: disable=W0613
-
         # TODO: Support Python3 range objects too.
         if python_version >= 300:
             return self, None, None
@@ -157,7 +197,8 @@ class ExpressionBuiltinRange1(ExpressionBuiltinRangeBase):
         low  = self.getLow()
 
         return self.computeBuiltinSpec(
-            given_values = (
+            constraint_collection = constraint_collection,
+            given_values          = (
                 low,
             )
         )
@@ -182,8 +223,6 @@ class ExpressionBuiltinRange1(ExpressionBuiltinRangeBase):
         if element_index > length:
             return None
 
-        from .NodeMakingHelpers import makeConstantReplacementNode
-
         # TODO: Make sure to cast element_index to what CPython will give, for
         # now a downcast will do.
         return makeConstantReplacementNode(
@@ -198,7 +237,7 @@ class ExpressionBuiltinRange1(ExpressionBuiltinRangeBase):
 class ExpressionBuiltinRange2(ExpressionBuiltinRangeBase):
     kind = "EXPRESSION_BUILTIN_RANGE2"
 
-    named_children = ( "low", "high" )
+    named_children = ("low", "high")
 
     def __init__(self, low, high, source_ref):
         ExpressionBuiltinRangeBase.__init__(
@@ -210,21 +249,25 @@ class ExpressionBuiltinRange2(ExpressionBuiltinRangeBase):
             source_ref = source_ref
         )
 
-    getLow  = ExpressionChildrenHavingBase.childGetter( "low" )
-    getHigh = ExpressionChildrenHavingBase.childGetter( "high" )
+    getLow  = ExpressionChildrenHavingBase.childGetter("low")
+    getHigh = ExpressionChildrenHavingBase.childGetter("high")
 
     builtin_spec = BuiltinOptimization.builtin_range_spec
 
     def computeExpression(self, constraint_collection):
-        # Children can tell all we need to know, pylint: disable=W0613
-
         if python_version >= 300:
             return self, None, None
 
         low  = self.getLow()
         high = self.getHigh()
 
-        return self.computeBuiltinSpec( ( low, high ) )
+        return self.computeBuiltinSpec(
+            constraint_collection = constraint_collection,
+            given_values          = (
+                low,
+                high
+            )
+        )
 
     def getIterationLength(self):
         low  = self.getLow()
@@ -240,7 +283,7 @@ class ExpressionBuiltinRange2(ExpressionBuiltinRangeBase):
         if high is None:
             return None
 
-        return max( 0, high - low )
+        return max(0, high - low)
 
     def canPredictIterationValues(self):
         return self.getIterationLength() is not None
@@ -264,8 +307,6 @@ class ExpressionBuiltinRange2(ExpressionBuiltinRangeBase):
         if result >= high:
             return None
         else:
-            from .NodeMakingHelpers import makeConstantReplacementNode
-
             return makeConstantReplacementNode(
                 constant = result,
                 node     = self
@@ -278,7 +319,11 @@ class ExpressionBuiltinRange2(ExpressionBuiltinRangeBase):
 class ExpressionBuiltinRange3(ExpressionBuiltinRangeBase):
     kind = "EXPRESSION_BUILTIN_RANGE3"
 
-    named_children = ( "low", "high", "step" )
+    named_children = (
+        "low",
+        "high",
+        "step"
+    )
 
     def __init__(self, low, high, step, source_ref):
         ExpressionBuiltinRangeBase.__init__(
@@ -291,15 +336,13 @@ class ExpressionBuiltinRange3(ExpressionBuiltinRangeBase):
             source_ref = source_ref
         )
 
-    getLow  = ExpressionChildrenHavingBase.childGetter( "low" )
-    getHigh = ExpressionChildrenHavingBase.childGetter( "high" )
-    getStep = ExpressionChildrenHavingBase.childGetter( "step" )
+    getLow  = ExpressionChildrenHavingBase.childGetter("low")
+    getHigh = ExpressionChildrenHavingBase.childGetter("high")
+    getStep = ExpressionChildrenHavingBase.childGetter("step")
 
     builtin_spec = BuiltinOptimization.builtin_range_spec
 
     def computeExpression(self, constraint_collection):
-        # Children can tell all we need to know, pylint: disable=W0613
-
         if python_version >= 300:
             return self, None, None
 
@@ -307,7 +350,14 @@ class ExpressionBuiltinRange3(ExpressionBuiltinRangeBase):
         high = self.getHigh()
         step = self.getStep()
 
-        return self.computeBuiltinSpec( ( low, high, step ) )
+        return self.computeBuiltinSpec(
+            constraint_collection = constraint_collection,
+            given_values          = (
+                low,
+                high,
+                step
+            )
+        )
 
     def getIterationLength(self):
         low  = self.getLow()
@@ -337,18 +387,18 @@ class ExpressionBuiltinRange3(ExpressionBuiltinRangeBase):
             if step < 0:
                 estimate = 0
             else:
-                estimate = math.ceil( float( high - low ) / step )
+                estimate = math.ceil(float(high - low) / step)
         else:
             if step > 0:
                 estimate = 0
             else:
-                estimate = math.ceil( float( high - low ) / step )
+                estimate = math.ceil(float(high - low) / step)
 
-        estimate = round( estimate )
+        estimate = round(estimate)
 
         assert not estimate < 0
 
-        return int( estimate )
+        return int(estimate)
 
     def canPredictIterationValues(self):
         return self.getIterationLength() is not None
@@ -371,8 +421,6 @@ class ExpressionBuiltinRange3(ExpressionBuiltinRangeBase):
         if result >= high:
             return None
         else:
-            from .NodeMakingHelpers import makeConstantReplacementNode
-
             return makeConstantReplacementNode(
                 constant = result,
                 node     = self
@@ -401,6 +449,6 @@ class ExpressionBuiltinXrange(ExpressionChildrenHavingBase):
     def computeExpression(self, constraint_collection):
         return self, None, None
 
-    getLow  = ExpressionChildrenHavingBase.childGetter( "low" )
-    getHigh = ExpressionChildrenHavingBase.childGetter( "high" )
-    getStep = ExpressionChildrenHavingBase.childGetter( "step" )
+    getLow  = ExpressionChildrenHavingBase.childGetter("low")
+    getHigh = ExpressionChildrenHavingBase.childGetter("high")
+    getStep = ExpressionChildrenHavingBase.childGetter("step")
