@@ -1,4 +1,4 @@
-#     Copyright 2014, Kay Hayen, mailto:kay.hayen@gmail.com
+#     Copyright 2015, Kay Hayen, mailto:kay.hayen@gmail.com
 #
 #     Part of "Nuitka", an optimizing Python compiler that is compatible and
 #     integrates with CPython, but also works on its own.
@@ -15,51 +15,61 @@
 #     See the License for the specific language governing permissions and
 #     limitations under the License.
 #
-""" Tree nodes for builtin references.
+""" Tree nodes for built-in references.
 
-There is 2 major types of builtin references. One is the values from
-builtins, the other is builtin exceptions. They work differently and
+There is 2 major types of built-in references. One is the values from
+built-ins, the other is built-in exceptions. They work differently and
 mean different things, but they have similar origin, that is, access
 to variables only ever read.
 
 """
 
 
-from .NodeBases import NodeBase, CompileTimeConstantExpressionMixin
-
-from .ConstantRefNodes import ExpressionConstantRef
-
-from nuitka.optimizations import BuiltinOptimization
-
 from nuitka.Builtins import (
+    builtin_anon_names,
     builtin_exception_names,
     builtin_exception_values,
-    builtin_anon_names,
     builtin_names
 )
+from nuitka.optimizations import BuiltinOptimization
+from nuitka.utils.Utils import python_version
+
+from .ConstantRefNodes import ExpressionConstantRef
+from .ExceptionNodes import ExpressionBuiltinMakeException
+from .NodeBases import CompileTimeConstantExpressionMixin, NodeBase
 
-from nuitka.Utils import python_version
 
 class ExpressionBuiltinRefBase(CompileTimeConstantExpressionMixin, NodeBase):
     def __init__(self, builtin_name, source_ref):
-        NodeBase.__init__( self, source_ref = source_ref )
-        CompileTimeConstantExpressionMixin.__init__( self )
+        NodeBase.__init__(self, source_ref = source_ref)
+        CompileTimeConstantExpressionMixin.__init__(self)
 
         self.builtin_name = builtin_name
 
-    def makeCloneAt(self, source_ref):
-        return self.__class__(self.builtin_name, source_ref)
-
     def getDetails(self):
-        return { "builtin_name" : self.builtin_name }
+        return {
+            "builtin_name" : self.builtin_name
+        }
 
     def getBuiltinName(self):
         return self.builtin_name
 
-    def mayHaveSideEffects(self):
-        # Referencing the builtin name has no side effect
+    def isKnownToBeHashable(self):
+        return True
+
+    def mayRaiseException(self, exception_type):
         return False
 
+    def mayHaveSideEffects(self):
+        return False
+
+    def getStrValue(self):
+        return ExpressionConstantRef(
+            constant      = str(self.getCompileTimeConstant()),
+            user_provided = True,
+            source_ref    = self.getSourceReference(),
+        )
+
 
 class ExpressionBuiltinRef(ExpressionBuiltinRefBase):
     kind = "EXPRESSION_BUILTIN_REF"
@@ -74,7 +84,6 @@ class ExpressionBuiltinRef(ExpressionBuiltinRefBase):
         )
 
     def isCompileTimeConstant(self):
-        # Virtual method, pylint: disable=R0201
         return True
 
     def getCompileTimeConstant(self):
@@ -96,45 +105,43 @@ class ExpressionBuiltinRef(ExpressionBuiltinRefBase):
             )
 
             return new_node, "new_constant", """\
-Builtin constant %s resolved""" % self.builtin_name
+Built-in constant '%s' resolved.""" % self.builtin_name
 
         return self, None, None
 
-    def computeExpressionCall(self, call_node, constraint_collection):
+    def computeExpressionCall(self, call_node, call_args, call_kw,
+                              constraint_collection):
         from nuitka.optimizations.OptimizeBuiltinCalls import computeBuiltinCall
 
+        # Anything may happen. On next pass, if replaced, we might be better
+        # but not now.
+        constraint_collection.onExceptionRaiseExit(BaseException)
+
         new_node, tags, message = computeBuiltinCall(
             call_node = call_node,
             called    = self
         )
 
-        if new_node.isExpressionBuiltinLocals() or \
-           new_node.isExpressionBuiltinEval():
-            constraint_collection.assumeUnclearLocals()
+        if self.builtin_name in ("dir", "eval", "exec", "execfile", "locals", "vars"):
+            # Just inform the collection that all has escaped.
+            constraint_collection.onLocalsUsage()
 
         return new_node, tags, message
 
     def getStringValue(self):
-        return repr( self.getCompileTimeConstant() )
+        return repr(self.getCompileTimeConstant())
 
     def isKnownToBeIterable(self, count):
         # TODO: Why yes, some may be, could be told here.
         return None
 
-    def mayProvideReference(self):
-        # Dedicated code returns which returns from builtin module dictionary, but isn't
-        # available for Python3 yet.
-
-        return python_version >= 300
-
 
 class ExpressionBuiltinOriginalRef(ExpressionBuiltinRef):
     kind = "EXPRESSION_BUILTIN_ORIGINAL_REF"
 
     def isCompileTimeConstant(self):
-        # TODO: Actually the base class should not be constant and this one should be.
-
-        # Virtual method, pylint: disable=R0201
+        # TODO: Actually the base class should not be constant and this
+        # one should be.
         return False
 
     def computeExpression(self, constraint_collection):
@@ -147,7 +154,7 @@ class ExpressionBuiltinAnonymousRef(ExpressionBuiltinRefBase):
     kind = "EXPRESSION_BUILTIN_ANONYMOUS_REF"
 
     def __init__(self, builtin_name, source_ref):
-        assert builtin_name in builtin_anon_names
+        assert builtin_name in builtin_anon_names, builtin_name
 
         ExpressionBuiltinRefBase.__init__(
             self,
@@ -156,15 +163,8 @@ class ExpressionBuiltinAnonymousRef(ExpressionBuiltinRefBase):
         )
 
     def isCompileTimeConstant(self):
-        # Virtual method, pylint: disable=R0201
         return True
 
-    def mayProvideReference(self):
-        # No reference provided from this, there are just a global identifiers, or
-        # accesses to them.
-
-        return False
-
     def getCompileTimeConstant(self):
         return builtin_anon_names[ self.builtin_name ]
 
@@ -172,7 +172,7 @@ class ExpressionBuiltinAnonymousRef(ExpressionBuiltinRefBase):
         return self, None, None
 
     def getStringValue(self):
-        return repr( self.getCompileTimeConstant() )
+        return repr(self.getCompileTimeConstant())
 
 
 class ExpressionBuiltinExceptionRef(ExpressionBuiltinRefBase):
@@ -188,41 +188,37 @@ class ExpressionBuiltinExceptionRef(ExpressionBuiltinRefBase):
         )
 
     def getDetails(self):
-        return { "exception_name" : self.builtin_name }
+        return {
+            "exception_name" : self.builtin_name
+        }
 
     getExceptionName = ExpressionBuiltinRefBase.getBuiltinName
 
     def isCompileTimeConstant(self):
-        # Virtual method, pylint: disable=R0201
         return True
 
-    def mayProvideReference(self):
-        # No reference provided from this, it's just a global identifier.
-
+    def mayRaiseException(self, exception_type):
         return False
 
     def getCompileTimeConstant(self):
-        return builtin_exception_values[ self.builtin_name ]
+        return builtin_exception_values[self.builtin_name]
 
     def computeExpression(self, constraint_collection):
-        # Children can tell all we need to know, pylint: disable=W0613
-
+        # Not much that can be done here.
         return self, None, None
 
-    def computeExpressionCall(self, call_node, constraint_collection):
+    def computeExpressionCall(self, call_node, call_args, call_kw,
+                              constraint_collection):
         exception_name = self.getExceptionName()
 
         # TODO: Keyword only arguments of it, are not properly handled yet by
         # the built-in call code.
         if exception_name == "ImportError" and python_version >= 330:
-            kw = call_node.getCallKw()
-
-            if not kw.isExpressionConstantRef() or kw.getConstant() != {}:
+            if call_kw is not None and \
+               (not call_kw.isExpressionConstantRef() or call_kw.getConstant() != {}):
                 return call_node, None, None
 
         def createBuiltinMakeException(args, source_ref):
-            from nuitka.nodes.ExceptionNodes import ExpressionBuiltinMakeException
-
             return ExpressionBuiltinMakeException(
                 exception_name = exception_name,
                 args           = args,
@@ -241,4 +237,4 @@ class ExpressionBuiltinExceptionRef(ExpressionBuiltinRefBase):
         if new_node is None:
             return call_node, None, None
 
-        return new_node, "new_expression", "detected builtin exception making"
+        return new_node, "new_expression", "Detected built-in exception making."
