@@ -1,4 +1,4 @@
-#     Copyright 2014, Kay Hayen, mailto:kay.hayen@gmail.com
+#     Copyright 2015, Kay Hayen, mailto:kay.hayen@gmail.com
 #
 #     Part of "Nuitka", an optimizing Python compiler that is compatible and
 #     integrates with CPython, but also works on its own.
@@ -15,77 +15,70 @@
 #     See the License for the specific language governing permissions and
 #     limitations under the License.
 #
-""" Reformulation of classes
+""" Reformulation of class statements.
 
-Consult the developmer manual for information. TODO: Add ability to sync
+Consult the developer manual for information. TODO: Add ability to sync
 source code comments with developer manual sections.
 
 """
 
-from nuitka.nodes.VariableRefNodes import (
+from nuitka.nodes.AssignNodes import (
     ExpressionTargetTempVariableRef,
     ExpressionTargetVariableRef,
-    ExpressionTempVariableRef,
-    ExpressionVariableRef
+    StatementAssignmentVariable,
+    StatementReleaseVariable
 )
-from nuitka.nodes.ConstantRefNodes import ExpressionConstantRef
-from nuitka.nodes.BuiltinRefNodes import ExpressionBuiltinRef
-from nuitka.nodes.ComparisonNodes import ExpressionComparison
-
-from nuitka.nodes.CallNodes import (
-    ExpressionCallNoKeywords,
-    ExpressionCall
-)
-from nuitka.nodes.TypeNodes import ExpressionBuiltinType1
 from nuitka.nodes.AttributeNodes import (
     ExpressionAttributeLookup,
     ExpressionBuiltinHasattr
 )
-from nuitka.nodes.SubscriptNodes import ExpressionSubscriptLookup
-from nuitka.nodes.FunctionNodes import (
-    ExpressionFunctionCreation,
-    ExpressionFunctionBody,
-    ExpressionFunctionCall,
-    ExpressionFunctionRef
-)
+from nuitka.nodes.BuiltinRefNodes import ExpressionBuiltinRef
+from nuitka.nodes.CallNodes import ExpressionCall, ExpressionCallNoKeywords
 from nuitka.nodes.ClassNodes import ExpressionSelectMetaclass
-from nuitka.nodes.ContainerMakingNodes import ExpressionMakeTuple
-from nuitka.nodes.ContainerOperationNodes import (
-    StatementDictOperationRemove,
-    ExpressionDictOperationGet
-)
-from nuitka.nodes.StatementNodes import StatementsSequence
-
+from nuitka.nodes.ComparisonNodes import ExpressionComparison
 from nuitka.nodes.ConditionalNodes import (
     ExpressionConditional,
     StatementConditional
 )
-from nuitka.nodes.ReturnNodes import StatementReturn
-from nuitka.nodes.AssignNodes import (
-    StatementAssignmentVariable,
-    StatementDelVariable
+from nuitka.nodes.ConstantRefNodes import ExpressionConstantRef
+from nuitka.nodes.ContainerMakingNodes import ExpressionMakeTuple
+from nuitka.nodes.ContainerOperationNodes import (
+    ExpressionDictOperationGet,
+    StatementDictOperationRemove
+)
+from nuitka.nodes.FunctionNodes import (
+    ExpressionClassBody,
+    ExpressionFunctionCall,
+    ExpressionFunctionCreation,
+    ExpressionFunctionQualnameRef,
+    ExpressionFunctionRef
 )
-
 from nuitka.nodes.GlobalsLocalsNodes import (
     ExpressionBuiltinLocals,
     StatementSetLocals
 )
-
 from nuitka.nodes.ParameterSpecs import ParameterSpec
+from nuitka.nodes.ReturnNodes import StatementReturn
+from nuitka.nodes.SubscriptNodes import ExpressionSubscriptLookup
+from nuitka.nodes.TypeNodes import ExpressionBuiltinType1
+from nuitka.nodes.VariableRefNodes import (
+    ExpressionTempVariableRef,
+    ExpressionVariableRef
+)
+from nuitka.utils import Utils
 
 from .Helpers import (
-    makeStatementsSequenceFromStatement,
-    makeSequenceCreationOrConstant,
-    makeDictCreationOrConstant,
-    makeStatementsSequence,
+    buildNode,
+    buildNodeList,
     buildStatementsNode,
     extractDocFromBody,
-    buildNodeList,
-    buildNode,
-    getKind
+    getKind,
+    makeDictCreationOrConstant,
+    makeSequenceCreationOrConstant,
+    makeStatementsSequence,
+    makeStatementsSequenceFromStatement
 )
-
-from nuitka import Utils
+from .ReformulationTryFinallyStatements import makeTryFinallyStatement
 
 # TODO: Once we start to modify these, we should make sure, the copy is not
 # shared.
@@ -130,7 +123,7 @@ def _buildClassNode3(provider, node, source_ref):
         name       = "prepared"
     )
 
-    class_creation_function = ExpressionFunctionBody(
+    class_creation_function = ExpressionClassBody(
         provider   = provider,
         is_class   = True,
         parameters = make_class_parameters,
@@ -139,10 +132,6 @@ def _buildClassNode3(provider, node, source_ref):
         source_ref = source_ref
     )
 
-    # Hack: This allows some APIs to work although this is not yet officially a
-    # child yet.
-    class_creation_function.parent = provider
-
     body = buildStatementsNode(
         provider   = class_creation_function,
         nodes      = class_statements,
@@ -151,16 +140,19 @@ def _buildClassNode3(provider, node, source_ref):
     )
 
     source_ref_orig = source_ref
-    source_ref = source_ref.atInternal()
 
     if body is not None:
         # The frame guard has nothing to tell its line number to.
         body.source_ref = source_ref
 
+    module_variable = class_creation_function.getVariableForAssignment(
+        "__module__"
+    )
+
     statements = [
         StatementSetLocals(
             new_locals = ExpressionTempVariableRef(
-                variable   = tmp_prepared.makeReference( provider ),
+                variable   = tmp_prepared,
                 source_ref = source_ref
             ),
             source_ref = source_ref
@@ -168,73 +160,100 @@ def _buildClassNode3(provider, node, source_ref):
         StatementAssignmentVariable(
             variable_ref = ExpressionTargetVariableRef(
                 variable_name = "__module__",
+                variable      = module_variable,
                 source_ref    = source_ref
             ),
-            source        = ExpressionConstantRef(
+            source       = ExpressionConstantRef(
                 constant      = provider.getParentModule().getFullName(),
                 source_ref    = source_ref,
                 user_provided = True
             ),
-            source_ref   = source_ref.atInternal()
+            source_ref   = source_ref
         )
     ]
 
     if class_doc is not None:
+        doc_variable = class_creation_function.getVariableForAssignment(
+            "__doc__"
+        )
+
         statements.append(
             StatementAssignmentVariable(
                 variable_ref = ExpressionTargetVariableRef(
                     variable_name = "__doc__",
+                    variable      = doc_variable,
                     source_ref    = source_ref
                 ),
-                source        = ExpressionConstantRef(
+                source       = ExpressionConstantRef(
                     constant      = class_doc,
                     source_ref    = source_ref,
                     user_provided = True
                 ),
-                source_ref   = source_ref.atInternal()
-            )
-        )
-
-    # The "__qualname__" attribute is new in Python 3.3.
-    if Utils.python_version >= 330:
-        qualname = class_creation_function.getFunctionQualname()
-
-        statements.append(
-            StatementAssignmentVariable(
-                variable_ref = ExpressionTargetVariableRef(
-                    variable_name = "__qualname__",
-                    source_ref    = source_ref
-                ),
-                source        = ExpressionConstantRef(
-                    constant      = qualname,
-                    source_ref    = source_ref,
-                    user_provided = True
-                ),
                 source_ref   = source_ref
             )
         )
 
-    if Utils.python_version >= 340:
+    # The "__qualname__" attribute is new in Python 3.3.
+    if Utils.python_version >= 330:
+        qualname = class_creation_function.getFunctionQualname()
+        qualname_variable = class_creation_function.getVariableForAssignment(
+            "__qualname__"
+        )
+
+        if Utils.python_version < 340:
+            qualname_ref = ExpressionConstantRef(
+                constant      = qualname,
+                source_ref    = source_ref,
+                user_provided = True
+            )
+        else:
+            qualname_ref = ExpressionFunctionQualnameRef(
+                function_body = class_creation_function,
+                source_ref    = source_ref,
+            )
+
+        statements.append(
+            StatementAssignmentVariable(
+                variable_ref = ExpressionTargetVariableRef(
+                    variable_name = "__qualname__",
+                    variable      = qualname_variable,
+                    source_ref    = source_ref
+                ),
+                source       = qualname_ref,
+                source_ref   = source_ref
+            )
+        )
+
+        if Utils.python_version >= 340:
+            qualname_assign = statements[-1]
+
+    if Utils.python_version >= 340 and False: # TODO: Temporarily reverted:
         tmp_class = class_creation_function.allocateTempVariable(
             temp_scope = None,
             name       = "__class__"
         )
 
         class_target_variable_ref = ExpressionTargetTempVariableRef(
-            variable   = tmp_class.makeReference(class_creation_function),
+            variable   = tmp_class,
             source_ref = source_ref
         )
         class_variable_ref = ExpressionTempVariableRef(
-            variable   = tmp_class.makeReference(class_creation_function),
+            variable   = tmp_class,
             source_ref = source_ref
         )
     else:
+        class_variable = class_creation_function.getVariableForAssignment(
+            "__class__"
+        )
+
         class_target_variable_ref = ExpressionTargetVariableRef(
             variable_name = "__class__",
+            variable      = class_variable,
             source_ref    = source_ref
         )
         class_variable_ref = ExpressionVariableRef(
             variable_name = "__class__",
+            variable      = class_variable,
             source_ref    = source_ref
         )
 
@@ -244,7 +263,7 @@ def _buildClassNode3(provider, node, source_ref):
             variable_ref = class_target_variable_ref,
             source       = ExpressionCall(
                 called     = ExpressionTempVariableRef(
-                    variable   = tmp_metaclass.makeReference(provider),
+                    variable   = tmp_metaclass,
                     source_ref = source_ref
                 ),
                 args       = makeSequenceCreationOrConstant(
@@ -256,7 +275,7 @@ def _buildClassNode3(provider, node, source_ref):
                             user_provided = True
                         ),
                         ExpressionTempVariableRef(
-                            variable   = tmp_bases.makeReference(provider),
+                            variable   = tmp_bases,
                             source_ref = source_ref
                         ),
                         ExpressionBuiltinLocals(
@@ -266,7 +285,7 @@ def _buildClassNode3(provider, node, source_ref):
                     source_ref    = source_ref
                 ),
                 kw         = ExpressionTempVariableRef(
-                    variable   = tmp_class_decl_dict.makeReference(provider),
+                    variable   = tmp_class_decl_dict,
                     source_ref = source_ref
                 ),
                 source_ref = source_ref
@@ -285,12 +304,12 @@ def _buildClassNode3(provider, node, source_ref):
         source_ref = source_ref
     )
 
-    # The class body is basically a function that implicitely, at the end
+    # The class body is basically a function that implicitly, at the end
     # returns its locals and cannot have other return statements contained.
 
     class_creation_function.setBody(body)
 
-    # The class body is basically a function that implicitely, at the end
+    # The class body is basically a function that implicitly, at the end
     # returns its created class and cannot have other return statements
     # contained.
 
@@ -311,13 +330,15 @@ def _buildClassNode3(provider, node, source_ref):
 
     for decorator in buildNodeList(
             provider,
-            reversed( node.decorator_list ),
+            reversed(node.decorator_list),
             source_ref
         ):
         decorated_body = ExpressionCallNoKeywords(
             called     = decorator,
             args       = ExpressionMakeTuple(
-                elements   = ( decorated_body, ),
+                elements   = (
+                    decorated_body,
+                ),
                 source_ref = source_ref
             ),
             source_ref = decorator.getSourceReference()
@@ -326,7 +347,7 @@ def _buildClassNode3(provider, node, source_ref):
     statements = (
         StatementAssignmentVariable(
             variable_ref = ExpressionTargetTempVariableRef(
-                variable   = tmp_bases.makeReference( provider ),
+                variable   = tmp_bases,
                 source_ref = source_ref
             ),
             source       = makeSequenceCreationOrConstant(
@@ -340,11 +361,11 @@ def _buildClassNode3(provider, node, source_ref):
         ),
         StatementAssignmentVariable(
             variable_ref = ExpressionTargetTempVariableRef(
-                variable   = tmp_class_decl_dict.makeReference(provider),
+                variable   = tmp_class_decl_dict,
                 source_ref = source_ref
             ),
             source       = makeDictCreationOrConstant(
-                keys      = [
+                keys       = [
                     ExpressionConstantRef(
                         constant      = keyword.arg,
                         source_ref    = source_ref,
@@ -353,24 +374,24 @@ def _buildClassNode3(provider, node, source_ref):
                     for keyword in
                     node.keywords
                 ],
-                values = [
-                    buildNode( provider, keyword.value, source_ref )
+                values     = [
+                    buildNode(provider, keyword.value, source_ref)
                     for keyword in
                     node.keywords
                 ],
                 lazy_order = False,
                 source_ref = source_ref
             ),
-            source_ref = source_ref
+            source_ref   = source_ref
         ),
         StatementAssignmentVariable(
             variable_ref = ExpressionTargetTempVariableRef(
-                variable   = tmp_metaclass.makeReference(provider),
+                variable   = tmp_metaclass,
                 source_ref = source_ref
             ),
             source       = ExpressionSelectMetaclass(
-                metaclass = ExpressionConditional(
-                    condition = ExpressionComparison(
+                metaclass  = ExpressionConditional(
+                    condition      = ExpressionComparison(
                         comparator = "In",
                         left       = ExpressionConstantRef(
                             constant      = "metaclass",
@@ -378,18 +399,14 @@ def _buildClassNode3(provider, node, source_ref):
                             user_provided = True
                         ),
                         right      = ExpressionTempVariableRef(
-                            variable   = tmp_class_decl_dict.makeReference(
-                                provider
-                            ),
+                            variable   = tmp_class_decl_dict,
                             source_ref = source_ref
                         ),
                         source_ref = source_ref
                     ),
-                    yes_expression = ExpressionDictOperationGet(
-                        dicte      = ExpressionTempVariableRef(
-                            variable   = tmp_class_decl_dict.makeReference(
-                                provider
-                            ),
+                    expression_yes = ExpressionDictOperationGet(
+                        dict_arg   = ExpressionTempVariableRef(
+                            variable   = tmp_class_decl_dict,
                             source_ref = source_ref
                         ),
                         key        = ExpressionConstantRef(
@@ -399,21 +416,19 @@ def _buildClassNode3(provider, node, source_ref):
                         ),
                         source_ref = source_ref
                     ),
-                    no_expression  = ExpressionConditional(
+                    expression_no  = ExpressionConditional(
                         condition      = ExpressionTempVariableRef(
-                            variable   = tmp_bases.makeReference( provider ),
+                            variable   = tmp_bases,
                             source_ref = source_ref
                         ),
-                        no_expression  = ExpressionBuiltinRef(
+                        expression_no  = ExpressionBuiltinRef(
                             builtin_name = "type",
                             source_ref   = source_ref
                         ),
-                        yes_expression = ExpressionBuiltinType1(
+                        expression_yes = ExpressionBuiltinType1(
                             value      = ExpressionSubscriptLookup(
-                                expression = ExpressionTempVariableRef(
-                                    variable   = tmp_bases.makeReference(
-                                        provider
-                                    ),
+                                subscribed = ExpressionTempVariableRef(
+                                    variable   = tmp_bases,
                                     source_ref = source_ref
                                 ),
                                 subscript  = ExpressionConstantRef(
@@ -429,13 +444,13 @@ def _buildClassNode3(provider, node, source_ref):
                     ),
                     source_ref     = source_ref
                 ),
-                bases     = ExpressionTempVariableRef(
-                    variable   = tmp_bases.makeReference( provider ),
+                bases      = ExpressionTempVariableRef(
+                    variable   = tmp_bases,
                     source_ref = source_ref
                 ),
                 source_ref = source_ref
             ),
-            source_ref = source_ref_orig
+            source_ref   = source_ref_orig
         ),
         StatementConditional(
             condition  = ExpressionComparison(
@@ -446,9 +461,7 @@ def _buildClassNode3(provider, node, source_ref):
                     user_provided = True
                 ),
                 right      = ExpressionTempVariableRef(
-                    variable   = tmp_class_decl_dict.makeReference(
-                        provider
-                    ),
+                    variable   = tmp_class_decl_dict,
                     source_ref = source_ref
                 ),
                 source_ref = source_ref
@@ -456,13 +469,11 @@ def _buildClassNode3(provider, node, source_ref):
             no_branch  = None,
             yes_branch = makeStatementsSequenceFromStatement(
                 statement = StatementDictOperationRemove(
-                    dicte = ExpressionTempVariableRef(
-                        variable   = tmp_class_decl_dict.makeReference(
-                            provider
-                        ),
+                    dict_arg   = ExpressionTempVariableRef(
+                        variable   = tmp_class_decl_dict,
                         source_ref = source_ref
                     ),
-                    key   = ExpressionConstantRef(
+                    key        = ExpressionConstantRef(
                         constant      = "metaclass",
                         source_ref    = source_ref,
                         user_provided = True
@@ -474,17 +485,13 @@ def _buildClassNode3(provider, node, source_ref):
         ),
         StatementAssignmentVariable(
             variable_ref = ExpressionTargetTempVariableRef(
-                variable   = tmp_prepared.makeReference(
-                    provider
-                ),
+                variable   = tmp_prepared,
                 source_ref = source_ref
             ),
             source       = ExpressionConditional(
-                condition = ExpressionBuiltinHasattr(
+                condition      = ExpressionBuiltinHasattr( # pylint: disable=E1120,E1123
                     object     = ExpressionTempVariableRef(
-                        variable   = tmp_metaclass.makeReference(
-                            provider
-                        ),
+                        variable   = tmp_metaclass,
                         source_ref = source_ref
                     ),
                     name       = ExpressionConstantRef(
@@ -494,17 +501,15 @@ def _buildClassNode3(provider, node, source_ref):
                     ),
                     source_ref = source_ref
                 ),
-                no_expression = ExpressionConstantRef(
+                expression_no  = ExpressionConstantRef(
                     constant      = {},
                     source_ref    = source_ref,
                     user_provided = True
                 ),
-                yes_expression = ExpressionCall(
+                expression_yes = ExpressionCall(
                     called     = ExpressionAttributeLookup(
-                        expression     = ExpressionTempVariableRef(
-                            variable   = tmp_metaclass.makeReference(
-                                provider
-                            ),
+                        source         = ExpressionTempVariableRef(
+                            variable   = tmp_metaclass,
                             source_ref = source_ref
                         ),
                         attribute_name = "__prepare__",
@@ -518,82 +523,74 @@ def _buildClassNode3(provider, node, source_ref):
                                 user_provided = True
                             ),
                             ExpressionTempVariableRef(
-                                variable   = tmp_bases.makeReference(
-                                    provider
-                                ),
+                                variable   = tmp_bases,
                                 source_ref = source_ref
                             )
                         ),
                         source_ref = source_ref
                     ),
                     kw         = ExpressionTempVariableRef(
-                        variable   = tmp_class_decl_dict.makeReference(
-                            provider
-                        ),
+                        variable   = tmp_class_decl_dict,
                         source_ref = source_ref
                     ),
                     source_ref = source_ref
                 ),
-                source_ref = source_ref
+                source_ref     = source_ref
             ),
-            source_ref = source_ref
+            source_ref   = source_ref
         ),
         StatementAssignmentVariable(
             variable_ref = ExpressionTargetVariableRef(
                 variable_name = node.name,
                 source_ref    = source_ref
             ),
-            source     = decorated_body,
+            source       = decorated_body,
+            source_ref   = source_ref
+        ),
+    )
+
+    if Utils.python_version >= 340:
+        class_assign = statements[-1]
+
+        # assert False, class_creation_function
+        class_creation_function.qualname_setup = class_assign, qualname_assign
+
+
+
+    final = (
+        StatementReleaseVariable(
+            variable   = tmp_bases,
             source_ref = source_ref
         ),
-        StatementDelVariable(
-            variable_ref = ExpressionTargetTempVariableRef(
-                variable   = tmp_bases.makeReference( provider ),
-                source_ref = source_ref
-            ),
-            tolerant   = False,
+        StatementReleaseVariable(
+            variable   = tmp_class_decl_dict,
             source_ref = source_ref
         ),
-        StatementDelVariable(
-            variable_ref = ExpressionTargetTempVariableRef(
-                variable   = tmp_class_decl_dict.makeReference(provider),
-                source_ref = source_ref
-            ),
-            tolerant   = False,
+        StatementReleaseVariable(
+            variable   = tmp_metaclass,
             source_ref = source_ref
         ),
-        StatementDelVariable(
-            variable_ref = ExpressionTargetTempVariableRef(
-                variable   = tmp_metaclass.makeReference(provider),
-                source_ref = source_ref
-            ),
-            tolerant   = False,
-            source_ref = source_ref
-        ),
-        StatementDelVariable(
-            variable_ref = ExpressionTargetTempVariableRef(
-                variable   = tmp_prepared.makeReference( provider ),
-                source_ref = source_ref
-            ),
-            tolerant   = False,
+        StatementReleaseVariable(
+            variable   = tmp_prepared,
             source_ref = source_ref
         )
     )
 
-    return StatementsSequence(
-        statements = statements,
+    return makeTryFinallyStatement(
+        provider   = provider,
+        tried      = statements,
+        final      = final,
         source_ref = source_ref
     )
 
 
-
 def _buildClassNode2(provider, node, source_ref):
     class_statements, class_doc = extractDocFromBody(node)
 
     # This function is the Python2 special case with special re-formulation as
     # according to developer manual.
 
-    function_body = ExpressionFunctionBody(
+    function_body = ExpressionClassBody(
         provider   = provider,
         is_class   = True,
         parameters = make_class_parameters,
@@ -613,7 +610,7 @@ def _buildClassNode2(provider, node, source_ref):
         # The frame guard has nothing to tell its line number to.
         body.source_ref = source_ref.atInternal()
 
-    # The class body is basically a function that implicitely, at the end
+    # The class body is basically a function that implicitly, at the end
     # returns its locals and cannot have other return statements contained, and
     # starts out with a variables "__module__" and potentially "__doc__" set.
     statements = [
@@ -622,7 +619,7 @@ def _buildClassNode2(provider, node, source_ref):
                 variable_name = "__module__",
                 source_ref    = source_ref
             ),
-            source        = ExpressionConstantRef(
+            source       = ExpressionConstantRef(
                 constant      = provider.getParentModule().getFullName(),
                 source_ref    = source_ref,
                 user_provided = True
@@ -638,7 +635,7 @@ def _buildClassNode2(provider, node, source_ref):
                     variable_name = "__doc__",
                     source_ref    = source_ref
                 ),
-                source        = ExpressionConstantRef(
+                source       = ExpressionConstantRef(
                     constant      = class_doc,
                     source_ref    = source_ref,
                     user_provided = True
@@ -663,14 +660,14 @@ def _buildClassNode2(provider, node, source_ref):
         source_ref = source_ref
     )
 
-    # The class body is basically a function that implicitely, at the end
+    # The class body is basically a function that implicitly, at the end
     # returns its locals and cannot have other return statements contained.
 
-    function_body.setBody( body )
+    function_body.setBody(body)
 
-    temp_scope = provider.allocateTempScope( "class_creation" )
+    temp_scope = provider.allocateTempScope("class_creation")
 
-    tmp_bases = provider.allocateTempVariable( temp_scope, "bases" )
+    tmp_bases = provider.allocateTempVariable(temp_scope, "bases")
     tmp_class_dict = provider.allocateTempVariable(temp_scope, "class_dict")
     tmp_metaclass = provider.allocateTempVariable(temp_scope, "metaclass")
     tmp_class = provider.allocateTempVariable(temp_scope, "class")
@@ -678,7 +675,7 @@ def _buildClassNode2(provider, node, source_ref):
     statements = [
         StatementAssignmentVariable(
             variable_ref = ExpressionTargetTempVariableRef(
-                variable   = tmp_bases.makeReference( provider ),
+                variable   = tmp_bases,
                 source_ref = source_ref
             ),
             source       = makeSequenceCreationOrConstant(
@@ -692,11 +689,11 @@ def _buildClassNode2(provider, node, source_ref):
         ),
         StatementAssignmentVariable(
             variable_ref = ExpressionTargetTempVariableRef(
-                variable   = tmp_class_dict.makeReference(provider),
+                variable   = tmp_class_dict,
                 source_ref = source_ref
             ),
             source       =   ExpressionFunctionCall(
-                function = ExpressionFunctionCreation(
+                function   = ExpressionFunctionCreation(
                     function_ref = ExpressionFunctionRef(
                         function_body = function_body,
                         source_ref    = source_ref
@@ -713,11 +710,11 @@ def _buildClassNode2(provider, node, source_ref):
         ),
         StatementAssignmentVariable(
             variable_ref = ExpressionTargetTempVariableRef(
-                variable   = tmp_metaclass.makeReference(provider),
+                variable   = tmp_metaclass,
                 source_ref = source_ref
             ),
             source       = ExpressionConditional(
-                condition =  ExpressionComparison(
+                condition      =  ExpressionComparison(
                     comparator = "In",
                     left       = ExpressionConstantRef(
                         constant      = "__metaclass__",
@@ -725,46 +722,46 @@ def _buildClassNode2(provider, node, source_ref):
                         user_provided = True
                     ),
                     right      = ExpressionTempVariableRef(
-                        variable   = tmp_class_dict.makeReference(provider),
+                        variable   = tmp_class_dict,
                         source_ref = source_ref
                     ),
                     source_ref = source_ref
                 ),
-                yes_expression = ExpressionDictOperationGet(
-                    dicte = ExpressionTempVariableRef(
-                        variable   = tmp_class_dict.makeReference(provider),
+                expression_yes = ExpressionDictOperationGet(
+                    dict_arg   = ExpressionTempVariableRef(
+                        variable   = tmp_class_dict,
                         source_ref = source_ref
                     ),
-                    key   = ExpressionConstantRef(
+                    key        = ExpressionConstantRef(
                         constant      = "__metaclass__",
                         source_ref    = source_ref,
                         user_provided = True
                     ),
                     source_ref = source_ref
                 ),
-                no_expression = ExpressionSelectMetaclass(
-                    metaclass = None,
-                    bases     = ExpressionTempVariableRef(
-                        variable   = tmp_bases.makeReference( provider ),
+                expression_no  = ExpressionSelectMetaclass(
+                    metaclass  = None,
+                    bases      = ExpressionTempVariableRef(
+                        variable   = tmp_bases,
                         source_ref = source_ref
                     ),
                     source_ref = source_ref
                 ),
-                source_ref = source_ref
+                source_ref     = source_ref
             ),
-            source_ref = source_ref
+            source_ref   = source_ref
         ),
         StatementAssignmentVariable(
             variable_ref = ExpressionTargetTempVariableRef(
-                variable   = tmp_class.makeReference(provider),
+                variable   = tmp_class,
                 source_ref = source_ref
             ),
-            source     = ExpressionCallNoKeywords(
-                called         = ExpressionTempVariableRef(
-                    variable   = tmp_metaclass.makeReference(provider),
+            source       = ExpressionCallNoKeywords(
+                called     = ExpressionTempVariableRef(
+                    variable   = tmp_metaclass,
                     source_ref = source_ref
                 ),
-                args           = ExpressionMakeTuple(
+                args       = ExpressionMakeTuple(
                     elements   = (
                         ExpressionConstantRef(
                             constant      = node.name,
@@ -772,13 +769,11 @@ def _buildClassNode2(provider, node, source_ref):
                             user_provided = True
                         ),
                         ExpressionTempVariableRef(
-                            variable   = tmp_bases.makeReference( provider ),
+                            variable   = tmp_bases,
                             source_ref = source_ref
                         ),
                         ExpressionTempVariableRef(
-                            variable   = tmp_class_dict.makeReference(
-                                provider
-                            ),
+                            variable   = tmp_class_dict,
                             source_ref = source_ref
                         )
                     ),
@@ -786,53 +781,27 @@ def _buildClassNode2(provider, node, source_ref):
                 ),
                 source_ref = source_ref
             ),
-            source_ref = source_ref
+            source_ref   = source_ref
         ),
-        StatementDelVariable(
-            variable_ref = ExpressionTargetTempVariableRef(
-                variable   = tmp_bases.makeReference( provider ),
-                source_ref = source_ref
-            ),
-            tolerant   = False,
-            source_ref = source_ref
-        ),
-        StatementDelVariable(
-            variable_ref = ExpressionTargetTempVariableRef(
-                variable   = tmp_class_dict.makeReference(provider),
-                source_ref = source_ref
-            ),
-            tolerant   = False,
-            source_ref = source_ref
-        ),
-        StatementDelVariable(
-            variable_ref = ExpressionTargetTempVariableRef(
-                variable   = tmp_metaclass.makeReference(provider),
-                source_ref = source_ref
-            ),
-            tolerant   = False,
-            source_ref = source_ref
-        )
     ]
 
     for decorator in buildNodeList(
             provider,
-            reversed( node.decorator_list ),
+            reversed(node.decorator_list),
             source_ref
         ):
         statements.append(
             StatementAssignmentVariable(
                 variable_ref = ExpressionTargetTempVariableRef(
-                    variable   = tmp_class.makeReference(provider),
+                    variable   = tmp_class,
                     source_ref = source_ref
                 ),
                 source       = ExpressionCallNoKeywords(
                     called     = decorator,
                     args       = ExpressionMakeTuple(
-                        elements  = (
+                        elements   = (
                             ExpressionTempVariableRef(
-                                variable   = tmp_class.makeReference(
-                                    provider
-                                ),
+                                variable   = tmp_class,
                                 source_ref = source_ref
                             ),
                         ),
@@ -850,37 +819,53 @@ def _buildClassNode2(provider, node, source_ref):
                 variable_name = node.name,
                 source_ref    = source_ref
             ),
-            source     = ExpressionTempVariableRef(
-                variable   = tmp_class.makeReference(provider),
+            source       = ExpressionTempVariableRef(
+                variable   = tmp_class,
                 source_ref = source_ref
             ),
+            source_ref   = source_ref
+        )
+    )
+
+    final = (
+        StatementReleaseVariable(
+            variable   = tmp_class,
+            source_ref = source_ref
+        ),
+        StatementReleaseVariable(
+            variable   = tmp_bases,
+            source_ref = source_ref
+        ),
+        StatementReleaseVariable(
+            variable   = tmp_class_dict,
+            source_ref = source_ref
+        ),
+        StatementReleaseVariable(
+            variable   = tmp_metaclass,
             source_ref = source_ref
         )
     )
 
-    statements.append(
-        StatementDelVariable(
-            variable_ref = ExpressionTargetTempVariableRef(
-                variable   = tmp_class.makeReference(provider),
-                source_ref = source_ref
-            ),
-            tolerant   = False,
-            source_ref = source_ref
-        )
-    )
-
-    return StatementsSequence(
-        statements = statements,
+    return makeTryFinallyStatement(
+        provider   = function_body,
+        tried      = statements,
+        final      = final,
         source_ref = source_ref
     )
 
+
 def buildClassNode(provider, node, source_ref):
-    assert getKind( node ) == "ClassDef"
+    assert getKind(node) == "ClassDef"
+
+    # There appears to be a inconsistency with the top level line number
+    # not being the one really the class has, if there are bases, and a
+    # decorator.
+    if node.bases:
+        source_ref = source_ref.atLineNumber(node.bases[-1].lineno)
 
     # Python2 and Python3 are similar, but fundamentally different, so handle
     # them in dedicated code.
-
-    if Utils.python_version >= 300:
-        return _buildClassNode3( provider, node, source_ref )
+    if Utils.python_version < 300:
+        return _buildClassNode2(provider, node, source_ref)
     else:
-        return _buildClassNode2( provider, node, source_ref )
+        return _buildClassNode3(provider, node, source_ref)
